package router

import (
    "time"
    "errors"
    "net/http"
    "nexus-server/users"
    "nexus-common/params"
    "golang.org/x/crypto/bcrypt"
    "github.com/golang-jwt/jwt"
    "github.com/labstack/echo/v4"
    "github.com/labstack/echo/v4/middleware"
    "github.com/go-playground/validator/v10"
    "github.com/sethvargo/go-password/password"
)

// jwt.StandardClaims is added as an embedded type to provide fields like user email and expiration time.
type customClaims struct {
    Email string
    jwt.StandardClaims
}

type auth struct {
    users *users.Users
    tokenAccess middleware.JWTConfig
}

var loginAttempts map[string]int  // Number of consecutive wrong logins for the same user
var lockedUsers map[string]time.Time   // Locked users and timestamps when their accounts were locked
var jwtSecretKey string

// Creates an auth object.
func NewAuth(u *users.Users) (*auth, error) {
    loginAttempts = make(map[string]int)
    lockedUsers = make(map[string]time.Time)

    // TODO: better to store jwtSecretKey in an env. variable?
    // Generates a 64 characters long key with 20 digits, 0 symbols,
    // allowing upper and lower case letters, and allowing repeat characters.
    secret, err := password.Generate(64, 20, 0, false, true)
    if err != nil {
        return nil, errors.New("Secret generation error: "+err.Error())
    }
    jwtSecretKey = secret

    return &auth {
        users : u,
        tokenAccess : middleware.JWTConfig {
            ContextKey: "user",
            SigningKey: []byte(jwtSecretKey),
            Claims:     &customClaims{},
            ErrorHandlerWithContext: func(err error, c echo.Context) error {
                return echo.NewHTTPError(http.StatusUnauthorized, "access denied")
            },
        },
    }, nil
}

// Token used by the JWT middleware to protect access to the specified group of routes.
func (auth *auth)GetTokenAccess() middleware.JWTConfig {
    return auth.tokenAccess
}

// curl --cacert ca.pem -X POST https://localhost:1077/login -H 'Content-Type: application/json' -d '{"email": "johndoe@nexus.org", "pwd":"the_user_password"}'
// Returns a jwt token if authentication was successful:
// {"token":"<jwt token>"}
func (auth *auth)Login(c echo.Context) error {
    // Deserializes the JSON body and checks its validity.
    reqUser := new(params.Login)
    if err := decodeJson(c, &reqUser); err != nil {
        return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
    }
    if err := validator.New(validator.WithRequiredStructEnabled()).Struct(reqUser); err != nil {
        return echo.NewHTTPError(http.StatusBadRequest, err.Error())
    }

    // User exists?
    user, err := auth.users.GetUserByEmail(reqUser.Email)
    if err != nil {
        // return echo.NewHTTPError(http.StatusUnauthorized, "User not found")
        return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
    }

    // Is user locked?
    lockedTime, found := lockedUsers[user.Email]
    if found {
        // If locked time has elapsed, unlock the user.
        t := time.Now()
        elapsedTime := t.Sub(lockedTime)
        if elapsedTime.Hours() > conf.Auth.LockedTimeInHours {
            if err := UnlockUser(c, user.Email); err != nil {
                return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
            }
        } else {
            log.Info("User ", user.Email ," is locked")
            return echo.NewHTTPError(http.StatusUnauthorized, "User is locked")
        }
    }

    // Checks the user password is correct.
    // If a user enters the wrong password too many times, the user will be locked for a number of hours.
    err = bcrypt.CompareHashAndPassword([]byte(user.Pwd), []byte(reqUser.Pwd))
    // If err == nil, passwords match!
    if (err == nil) {
        loginAttempts[user.Email] = 0
    } else {
        // Stores the number of attempted logins.
        count, found := loginAttempts[user.Email]
        if found {
            count = count+1
        } else {
            count = 1
        }
        loginAttempts[user.Email] = count
        log.Info("Login attempt ", count, " for user ", user.Email)
        if count > conf.Auth.MaxLoginAttempts {
            // Locks user
            lockedUsers[user.Email] = time.Now()
            log.Info("User ", user.Email ," is locked: too many wrong password attempts!")
            return echo.NewHTTPError(http.StatusUnauthorized, "User is locked")
        }

        // If passwords don't match, return a 401 status.
        return echo.NewHTTPError(http.StatusUnauthorized, "Invalid password")
    }

    tok, err := createToken(user.Email)
    if err != nil {
        return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
    }

    tokenSerialized := params.Token{ Token: tok }
    return c.JSON(http.StatusOK, tokenSerialized)
}

// Creates a JWT token and claims with the expiration time for the given user.
func createToken(userEmail string) (string, error) {
    expirationTime := time.Now().Add(time.Duration(conf.Auth.AccessTokenExpirationTimeInHours)*time.Hour)

    // Creates the claims
    claims := &customClaims{
        Email: userEmail,
        StandardClaims: jwt.StandardClaims {
            // JWT expresses expiration time in Unix milliseconds
            ExpiresAt: expirationTime.Unix(),
        },
    }

    // Declares the token with the algorithm used for signing, and the claims.
    token := jwt.NewWithClaims(jwt.SigningMethodHS256, claims)

    // Signs the token and gets a string version of the token.
    secret := []byte(jwtSecretKey)
    tokenString, err := token.SignedString(secret)
    if err != nil {
        return "", err
    }

    return tokenString, nil
}

// Returns user email from the Echo context.
// If an error occured, returns the error.
func GetUserFromContext(c echo.Context) (string, error) {
    if c.Get("user") == nil {
        return "", errors.New("user not authenticated")
    }

    user := c.Get("user").(*jwt.Token)
    claims := user.Claims.(*customClaims)

    return claims.Email, nil
}

func UnlockUser(c echo.Context, emailToUnlock string) error {
    userEmail, err := GetUserFromContext(c)
    if err != nil {
        return err
    }

    _, found := lockedUsers[emailToUnlock]
    if found {
        // unlock user by removing it from the locked map
        delete(lockedUsers, emailToUnlock)
        loginAttempts[emailToUnlock] = 0  // reset the number of attempts
        log.Info("User ", userEmail, " unlocked user ", emailToUnlock)
        return nil
    } else {
        return errors.New("User not locked")
    }

}

/*
// Example of middleware
// Such a middleware could be used to handle a logout feature.
// We would keep a map of invalid tokens (the ones that logged out), and deny access to such tokens.
// Question: Entries in the map would have to be cleared at times to prevent it from growing too much.
//           This could be done when the token has been in the map for longer than its expiration time.

usersGroup.Use(auth.OnlyJoe)  // apply the middleware (to be added in router.go)

func OnlyJoe(next echo.HandlerFunc) echo.HandlerFunc {
    return func(c echo.Context) error {
        // Gets user token from context.
        if c.Get("user") == nil {
            return echo.NewHTTPError(http.StatusUnauthorized, "invalid user token\n")
        }
        user := c.Get("user").(*jwt.Token)
        claims := user.Claims.(*customClaims)
        username := claims.Username
        if username != "joe" {
            return echo.ErrUnauthorized
        }
        return next(c)
    }
}
*/
